import { Chevron, linkIcon } from "../utils/icons.js";
import {sleep, bs, isElement} from "../utils/lib.js";

const validate = (items, ids = []) => {
    const _str = obj => JSON.stringify(obj);
    if (!(items && Array.isArray(items)))
        throw Error('Invalid Menu Item. Must be an Array')
    if (items.length === 0)
        return console.warn('You need at leat 1 menu item to create Collapsible Manu. None provided')
    for (let item of items) {
        let id =item.id;
        if(id){
            if(typeof id !== 'string' || !id.match(/^[_a-z]/))
               throw Error('\'id\' is required for every field and must start with a letter or underscore (_). Trace: ' + _str(item));
            if(ids.includes(id))
               throw Error('Same \'id\' have been detected between collapsible items. Please use unique \'id\'. Trace: ' + _str(item))
            ids.push(id)
        }
        if (typeof item !== 'object')
            throw Error('Invalid Menu Item found.')
        const keys = Object.keys(item);
        if (keys.includes('children') && !keys.includes('id'))
            throw Error('All menu items with children must have a unique id property, Trace: ' + _str(item))
        if (!keys.includes('label'))
            throw Error('All menu items must have a label property, Trace: ' + _str(item))
        if (!keys.includes('children') && !keys.includes('href'))
            throw Error('All menu items without children must have href property, Trace: ' + _str(item))
        if (keys.includes('children') && (!Array.isArray(item.children) || item.children.length == 0))
            throw Error('children property must be an array with at least 1 value, Trace: ' + _str(item))
        if (keys.includes('children'))
            validate(item.children, ids);
    }
}

export default function(_, items, props){
    const { df, aic } = bs;
    let {grid, active, linkIcon: li, activeClass, onclick } = props;
    grid = typeof grid === 'string' && ['vertical', 'both', 'none'].includes(grid) ? grid : 'both';
    let _active = null;
    activeClass = activeClass ? activeClass : 'text-primary';
    li = li ? li : _(linkIcon());
    let instanceID = `collap_${helper.random(700000, 5677788) + Date.now()}_se`;
    let container = _(`< id = "${instanceID}"/>`);

    validate(items);
    const col_btn = (label, collapse_id) => {
        let btn = _('<button type="button"/>').style({
            whiteSpace: 'normal'
        }).addClass(['dropdown-item has-dropdown', df, aic]).addChild(label);
        collapse_id && btn.attr({
            'data-bs-toggle': 'collapse',
            'data-bs-target': `#${collapse_id}`,
        })
        return btn;
    }
    const left_line = () => _('</>').style({
        width: '16px',
        height: '2px',
        background: 'rgba(0,0,0,0.2)',
        marginLeft: '-16px',
        marginBottom: 'auto',
        marginTop: '12px',
        flexShrink: '0'
    });

    const activateClass = _a => {
        if (_active) {
            _active.removeClass(activeClass);
        }
        _a.addClass(activeClass);
        _active = _a;
    }

    const activateCollapse = async(id) => {
        await sleep(100);
        let target = _(`#${instanceID} a[href="${id}"]`);
        if (!target) return;
        let parent = target.parent();
        const _click = () => {
            let prev = parent.siblings().get(-1);
            if (parent.hasClass('jql-nested-collapse-container') &&
                !parent.hasClass('show') && isElement(prev)) {
                prev.click();
                parent = prev.parent();
                _click();
            }
        }
        activateClass(target);
        _click();
    }

    const add = (item, line) => {
        const children = item.children;
        let open = false;
        if (children) {

            const id = item.id
            const col = _('< class = "collapse jql-nested-collapse-container"/>')
                .id(id)
                .addChild(children.map(child => add(child, true)))
                .style({ marginLeft: '23px' });
            if (grid != 'none')
                col.style({ borderLeft: '2px solid rgba(0,0,0,0.2)' })
            const toggle = (btn, show) => {
                let style = {
                    transform: 'rotate(0deg)',
                    marginRight: ''
                }
                if (col.hasClass('show') || (show && open))
                    style = {
                        transform: 'rotate(90deg)',
                        marginRight: '2px'
                    }
                btn.children().get(line && grid === 'both' ? 1 : 0).style(style);
            }
            const btn = col_btn([line && grid === 'both' ? left_line() : '', Chevron(), item.label], id).on('click', async() => {
                open = !open;
                toggle(btn, true);
                await sleep(360)
                toggle(btn);
            });
            return [btn, col];
        } else {
            let cont = [];
            if (line && grid === 'both')
                cont.push(left_line())
            cont.push(li.clone().style({
                marginRight: '2px',
                flexShrink: '0',
                marginBottom: 'auto',
                marginTop: '6px'
            }));
            cont.push(item.label);
            let _a = _('<a/>').style({
                whiteSpace: 'normal'
            });
            return _a.attr({
                href: item.href
            }).on('click', () => {
                if (onclick && typeof onclick === 'function')
                    onclick({ href: item.href, target: _a.plain(0) })
                activateClass(_a);
            }).addClass(['dropdown-item no-dropdown', df, aic]).addChild(cont);
        }
    }

    items.forEach(item => {
        container.addClass('jql-nested-collapse-container').addChild(add(item))
    });

    active && activateCollapse(active)

    return { container, active: href => href ? activateCollapse(href) : _active };
}