
/**
 *  Has No dependencies
 */

import { create_el, isJQL, _object as _obj, _boolean, _primitive, isDocOrWin, docEl, docWinProp, _object, _array, _string, isElement, toElement, _number, oneArrayElements } from './utils/lib.js';
import handleMove from './self/methods/move.js';
import mountAt from './self/methods/pos-at.js';
import gentleScroll from './self/methods/gentle-scroll.js';
import { addClass, className, hasClass, removeClass } from './self/methods/classes.js';
import { addChild, appendTo, prependTo, html, insertAfter, insertBefore, parent, text } from './self/methods/nodes.js';
import { on, _on } from './self/methods/events.js';
import { attr, rmAttr, style, hasAttr } from './self/methods/attributes.js';
import Window from './self/methods/window.js';

/**
 *  Depends on Bootstrap 5
 */

import expandable from './bs/expandable.js';
import Form from './bs/form.js';
import Modal from './bs/modal.js';
import Carousel from './bs/carousel.js';
import OffCanvas from './bs/offcanvas.js';
import Card from './bs/card.js';
import Section from './bs/section.js';
import Alert from './bs/alert.js';
import Confirm from './bs/confirm.js';
import Prompt from './bs/prompt.js';
import Toast from './bs/toast.js';
import Collapse from './bs/collapse.js';
import Accordion from './bs/accordion.js';
import nestedCollapse from './bs/nested-collapse.js';


class Primitive{}
class JQL{}
/**
 * ----------------------------------------------------------------
 * jQuery Lite v1.0.0
 * ----------------------------------------------------------------
 * Write Less, do more!!!
 * @param {String | Element | Window | Document}  Element_or_Query_or_Tag_or_Window_or_Document 1) The tag for the new element you
 * you want to create in the form '&lt;tagName /&gt;'. You can add attributes same way as in HTML but must be a string entirely 
 * like '&lt;tagName style="background:red;color:pink" id="idname" /&gt;'.
 * 2) An Element created or picked from the DOM using JavaScript.
 * 3) The window or document object for just a few methods
 * @see https://www.247-dev.com/projects/jquerylite/doc#howtouse
 */

export default function jQueryLite(Element_or_Query_or_Tag_or_Window_or_Document) {
    let el = Element_or_Query_or_Tag_or_Window_or_Document;
    el = create_el(Element_or_Query_or_Tag_or_Window_or_Document);
    let allEl = oneArrayElements(el);

    const queryMethods = (collection, siblingsFor, onlyElements) =>{
        collection = [].slice.call(collection);
        let target = siblingsFor ? [siblingsFor.parentElement] : allEl;

        return {
        /**
         * Returns the results of a query. But also a particular result by index n. If you are accessing get
         * from siblings, that is el.siblings().get(n), n will be considered as sibling. -n for siblings behind
         * while +n for siblings infront.
         * @param {Number} n target result index
         * @see https://www.247-dev.com/projects/jquerylite/doc/get
         */  
            get(n) {
                let result;
                if(siblingsFor && typeof n === 'number'){
                    let thisIndex = collection.indexOf(siblingsFor);
                    let sum = thisIndex + n;
                    if (n === -1)
                        result = jQueryLite(siblingsFor).previousSibling(onlyElements);
                    if (n === 1)
                        result = jQueryLite(siblingsFor).nextSibling(onlyElements);
                    if (sum >= 0 && sum < collection.length)
                        result = collection[sum];
                }else
                   result =  typeof n === 'number' ? collection[n] : collection
                return jQueryLite(result)
            },

            /**
             * Filters the collection of element. Accepts a function, string and number, please 
             * the link attached
             * @param {*} exclude What to exclude
             * @see https://www.247-dev.com/projects/jquerylite/doc/exclude
             */

            except(exclude){
              let forChecking = [];
              let checked = [];
              exclude = _array(exclude, [exclude]);
    
              exclude.map(ex => {
                  if(_string(ex) && !checked.includes(ex))
                     forChecking = [...forChecking, ...jQueryLite(target).query(ex).get().plain()];
                  else if(isElement(ex))
                     forChecking = [...forChecking, ex];
                  checked.push(ex);
              });
    
              forChecking = forChecking.map(toElement);
              let excludedIndexes = exclude.filter(xc => typeof xc === 'number');
              let resultSoFar = collection.filter((c, index) => !excludedIndexes.includes(index));

              resultSoFar = resultSoFar.filter(c => !forChecking.includes(c));
              let callbacks = exclude.filter(ex => typeof ex === 'function');

              if(callbacks.length > 0)
                  resultSoFar = resultSoFar.filter(rsf => !callbacks.every(cb => cb(jQueryLite(rsf))))
                
              return queryMethods(resultSoFar, siblingsFor, onlyElements);
            }
        }
    }

    return {

        /**
         * Adds class(es) to element
         * @param {String | String[]} className className or array of classNames to be added. Can also be a string of 
         * whitespace separated classNames
         * @see https://www.247-dev.com/projects/jquerylite/doc#addclass
         */

        addClass(className){
            return jQueryLite(allEl.map(e => addClass(e, className)));
        },

        /**
         * Adds a class if absent or removes it if present
         * @param {String} className
         * @see https://www.247-dev.com/projects/jquerylite/doc#toggleclass
         */

        toggleClass(className){
            allEl.map(el => el.classList.toggle(className));
            return jQueryLite(allEl);
        },

        /**
         * Removes class(es) from an element
         * @param {String} className className or array of classNames to be added. Can also be a string of 
         * whitespace separated classNames
         * @see https://www.247-dev.com/projects/jquerylite/doc#removeclass
         */

        removeClass(className){
            
            return jQueryLite(allEl.map(el => removeClass(el, className)));
        },

        /**
         * Returns classname if value is not provided else overrides the existing class with value
         * @param {String | void} value 
         * @see https://www.247-dev.com/projects/jquerylite/doc#classname
         */

        className(value){
            return className(allEl[0], value);
        },

        /**
         * Returns true if the className(s) are all present in the element else false
         *  @param {String | String[]} className className or array of classNames to be added. Can also be a string of 
         * whitespace separated classNames
         * @see https://www.247-dev.com/projects/jquerylite/doc#hasclass
         */

        hasClass(className){
            return hasClass(allEl[0], className);
        },

        /**
         * Tells if the node is the first childNode or firstElementChild depending on the value for onlyElement
         * @param {Boolean} onlyElement By default, onlyElement is true, which would use Elements only. Else would use all Nodes
         * @see https://www.247-dev.com/projects/jquerylite/doc#isfirstchild
         */

        isFirstChild(onlyElement = true){
            return allEl[0].parentElement[onlyElement ? 'firstElementChild' : 'firstChild'] == allEl[0];
        },

        /**
         * Tells if the node is the last childNode or lastElementChild depending on the value for onlyElement
         * @param {Boolean} onlyElement By default, onlyElement is true, which would use Elements only. Else would use all Nodes
         * @see https://www.247-dev.com/projects/jquerylite/doc#islastchild
         */

        isLastChild(onlyElement = true){
            return allEl[0].parentElement[onlyElement ? 'lastElementChild' : 'lastChild'] == allEl[0];
        },
        index(){
            return isElement(allEl[0]) ? jQueryLite(allEl[0]).siblings().get().plain().indexOf(allEl[0]) : null;
        },

        /**
         * Returns the first childNode or firstElementChild depending on the value for onlyElement
         * @param {Boolean} onlyElement By default, onlyElement is true, which would use Elements only. Else would use all Nodes
         * @see https://www.247-dev.com/projects/jquerylite/doc#firstchild
         */

        firstChild(onlyElement = true){
            return jQueryLite(onlyElement ? allEl[0].firstElementChild : allEl[0].firstChild);
        },

        /**
         * Returns the last childNode or lastElementChild depending on the value for onlyElement
         * @param {Boolean} onlyElement By default, onlyElement is true, which would use Elements only. Else would use all Nodes
         * @see https://www.247-dev.com/projects/jquerylite/doc#lastchild
         */

        lastChild(onlyElement = true){
            return jQueryLite(onlyElement ? allEl[0].lastElementChild : allEl[0].lastChild)
        },

        /**
         * Returns the next adjecent childNode or Element depending on the value for onlyElement
         * @param {Boolean} onlyElement By default, onlyElement is true, which would use Elements only. Else would use all Nodes
         * @see https://www.247-dev.com/projects/jquerylite/doc#nextsibling
         */

        nextSibling(onlyElement = true){
            return jQueryLite(onlyElement ? allEl[0].nextElementSibling : allEl[0].nextSibling);
        },

        /**
         * Returns the previous adjecent childNode or Element depending on the value for onlyElement
         * @param {Boolean} onlyElement By default, onlyElement is true, which would use Elements only. Else would use all Nodes
         * @see https://www.247-dev.com/projects/jquerylite/doc#previoussibling
         */

        previousSibling(onlyElement = true){
            return jQueryLite(onlyElement ? allEl[0].previousElementSibling : allEl[0].previousSibling)
        },

        /**
         * Returns the id of the Element if id parament is void else would override the existing id with the id supplied
         * @param {Sring | void} id 
         * @see https://www.247-dev.com/projects/jquerylite/doc#id
         */

        id(id){
            return id && jQueryLite(allEl[0]).attr({ id }) || allEl[0].id
        },

        /**
         * For appending new Nodes
         * @param {String | Element | JQL | [String | Element | JQL]} content Pass a single child content 
         * or an array of child contents
         * @see https://www.247-dev.com/projects/jquerylite/doc#addchild
         */

        addChild(content){
           return jQueryLite(addChild(jQueryLite, allEl[0], content));
        },

        /**
         * Addes a new content before an existing child element in the DOM
         * @param {String | Element | JQL | [String | Element | JQL]} newContent Pass a single child content 
         * or an array of child contents
         * @param {Element | JQL} existing A child element already existing in the DOM before which the new content 
         * should be added.
         * @see https://www.247-dev.com/projects/jquerylite/doc#insertbefore
         */

        insertBefore(newContent, existing) {
           return jQueryLite(insertBefore(jQueryLite, allEl[0], newContent, existing));
        },

        /**
         * Addes a new content after an existing child element in the DOM
         * @param {String | Element | JQL | [String | Element | JQL]} newContent Pass a single child content 
         * or an array of child contents
         * @param {Element | JQL} existing A child element already existing in the DOM after which the new content 
         * should be added.
         * @see https://www.247-dev.com/projects/jquerylite/doc#insertafter
         */

        insertAfter(newContent, existing){
           return jQueryLite(insertAfter(jQueryLite, allEl[0], newContent, existing));
        },

        /**
         * If a string is provided, assumes it's attribute name and returns the value. Else if a valid object 
         * is provided, new attibutes are creted from it and the elemnt returned as JQL
         * @param {String | {}} keyValuePairs 
         * @see https://www.247-dev.com/projects/jquerylite/doc#attr
         */

        attr(keyValuePairs){
            let att = allEl.map(el => attr(el, keyValuePairs));
            return att.every(at => _object(at, null)) ? jQueryLite(att) : att[0];
        },

        /**
         * Removes attributes from an element
          @param {String | String[]} attrName attribute or array of attributes to be removed. Can also be a string of 
         * whitespace separated attributes
         */

        removeAttribute(attrName){
            return allEl.map(el => rmAttr(el, attrName));
        },
        
        /**
         * Returns true if attibutes are all present in an element else false.
          @param {String | String[]} attrName attribute or array of attributes to be checked. Can also be a string of 
         * whitespace separated attributes
         */

        hasAttribute(attrName){
            return hasAttr(allEl[0], attrName)
        },

        /**
         Adds or replace element styles
         * @param {String | {}} keyValuePairs If string is provided, would override the existing style with the provided one which should be in css format.
         * Else if valid object is provided, new styles are going to be added to the element from the object.
         */

        style(keyValuePairs){
            return jQueryLite(allEl.map(el => style(el, keyValuePairs)))
        },

        /**
         * Remove an element from the DOM
         */

        drop(){
            return allEl.map(el => el.parentElement.removeChild(el));
        },
        
        /**
         * Remove all childNodes of an element from the DOM
         */

        truncate(){
            allEl.map(el => el.innerHTML = null);
            return jQueryLite(allEl);
        },

        /**
         * Truncates and element and add new contents
         * @param {String | Element | JQL | [String | Element | JQL]} newContent Pass a single child content 
         * or an array of new contents
         */

        refill(newContent){
            return jQueryLite(allEl).truncate().addChild(newContent)
        },

        /**
         * Disables or enables an element depending on parameter
         * @param {Boolean} disable if true, will disable the element. Else, will enable it. Value is true by default
         */

        disable(disable){
            allEl.map(el => el.disabled = _boolean(disable));
            return jQueryLite(allEl);
        },

        /**
         * Clones an element
         * @param {Boolean} deep if true, will clone the element including it's childNodes else will clone only the 
         * element. true by default.
         */

        clone(deep){
            return jQueryLite(allEl.map(el => el.cloneNode(_boolean(deep))))
        },

        /**
         * Will click or listen for click event depending on the parameter
         * @param {Function | void} fn if function is set, an eventListener will be added with the function as callback 
         * else a click will be performed on the element.
         * @returns 
         */

        click(fn){
            allEl.map(el => typeof fn === 'function' ? jQueryLite(el).on('click', fn) : el.click())
            return jQueryLite(allEl);
        },

        /**
         * Usefull for form element only, to clear all fields.
         */

        reset(){
            allEl.map(el => el.reset());
            return jQueryLite(el);
        },

        /**
         * ----------------------------------------------------
         * Add EventListerner
         * ----------------------------------------------------
         * @param {String} event
         * @param {Function} callback
         * @param {Boolean | {}} useCapture
         * @see https://www.247-dev.com/projects/jquerylite/doc#on
         */

        on(event, callback, useCapture){
            return jQueryLite(allEl.map(el => on(jQueryLite, el, event, callback, useCapture)));
        },

        /**
         * ----------------------------------------------------
         * Remove EventListerner
         * ----------------------------------------------------
         * @param {String} event
         * @param {Function} callback
         * @param {Boolean | {}} useCapture
         * @see https://www.247-dev.com/projects/jquerylite/doc#_on
         */

        _on(event, callback, useCapture){
            return jQueryLite(allEl.map(el => _on(jQueryLite, el, event, callback, useCapture)));
        },

        /**
         * ------------------------------------------------------------------
         * Will make an element to be moved on the screen to any position by 
         * dragging from anywhere or a particular spot
         * ------------------------------------------------------------------
         * @param {{ 
         * movepoint: HTMLElement | jQL,
         * onMove: Function, 
         * onMoving: Function, 
         * onMoved: Function
         * }} props
         * 
         * @property movepoint Can only be moved from this point if set
         * @property onMove Called immidiately the element is moved
         * @property onMoving Called when the element is moving
         * @property onMoved Called when the element stops after moving
         * @condition The element must have a position property set to 'absolute' or 'fixed'. 
         * You have to set this value by yourself since the method won't temper with your styling
         * @see https://www.247-dev.com/projects/jquerylite/doc#movable
         */

        movable(props){
            return jQueryLite(allEl.map(el => handleMove(jQueryLite, {el, ..._obj(props) })));
        },

        /**
         * ----------------------------------------------------------------
         * Mount an element at particular point (x, y) on the viewport extracted 
         * from the Event Object where where it occured. x and y can manually 
         * be passed as well.
         * ----------------------------------------------------------------
         * @param {{ 
         * x: Number,
         * y: Number, 
         * } | Event} eventObject You can pass an Event Object so the element can be 
         * placed exactly where the event occured and it would only respond to clicks
         * outside the element. But for a particular position, then pass x and y 
         * properties as eventObject. You can also make the element movable
         * after mounting by passing movable properperties as well.
         * 
         * @param {{ 
         * movepoint: HTMLElement | jQL,
         * onMove: Function, 
         * onMoving: Function, 
         * onMoved: Function
         * height: Number,
         * width: Number
         * }} props
         * 
         * @property movepoint Can only be moved from this point if set
         * @property onMove Called immidiately the element is moved
         * @property onMoving Called when the element is moving
         * @property onMoved Called when the element stops after moving.
         * @see https://www.247-dev.com/projects/jquerylite/doc#mountat
         */
        mountAt(eventObject, props){
            return jQueryLite(allEl.map(el => mountAt(jQueryLite, {el, e: _obj(eventObject), ..._obj(props) })))
        },

        /**
         * ----------------------------------------------------------------
         * Makes an element collapsible
         * ----------------------------------------------------------------
         * @param {Boolean} toggle If true, will show the element initially. Default is false
         * @see https://www.247-dev.com/projects/jquerylite/doc/collapse
         */

        collapse(toggle){
            return Collapse(jQueryLite(allEl[0]), toggle)
        },

        /**
         * ----------------------------------------------------------------
         * Adds a Bootstrap 5 accordion to the target element
         * ----------------------------------------------------------------
         * @param {Element | Text} body Content of the accorfion
         * 
         * @param {{ 
         * title: Text,
         * toggle: Boolean
         * }} props
         * @see https://www.247-dev.com/projects/jquerylite/doc/accordion
         */

        accordion(body, props){
            return Accordion(jQueryLite, body, { container: allEl[0], ..._obj(props) });
        },

        /**
         * ----------------------------------------------------------------
         * Returns the carousel instance or creates a new one, use this if you have typed out the carousel 
         * elements by yourself. Else use the static method. _.carousel({...}) for a fresh carousel and reduced work.
         * ----------------------------------------------------------------
         * @param {Number} n Where to slide the carousel to immidiately the instance is ready. Default is 0, and this will start the carousel 
         * for cycling
         * @see https://www.247-dev.com/projects/jquerylite/doc/carousel-get-or-create
         */

        carousel(n = 0){
            return Carousel(jQueryLite, null, typeof n != 'number' ? 0 : n, allEl[0]);
        },

        /**
         * ----------------------------------------------------------------
         * Copies and zoom out the target element for more visibility if a user clicks on it
         * ----------------------------------------------------------------
         * @see https://www.247-dev.com/projects/jquerylite/doc/expandable
         */

        expandable(){
            return allEl.map(el => expandable(jQueryLite, el));
        },

        /**
         * ----------------------------------------------------------------
         * Animates the children of this element if the element is scrolled
         * ----------------------------------------------------------------
         * 
         * @param {{ 
         * transitionDuration: Number,
         * transitionDistance: Number, 
         * targets: String, 
         * transitionType: 'to-top' | 'to-bottom' | 'to-left' | 'to-right'
         * }} props
         * @see https://www.247-dev.com/projects/jquerylite/doc/gentlescroll
         */

        gentleScroll(props){
            return allEl.map(el => gentleScroll({ el: jQueryLite(el), ..._obj(props) }));
        },

        /**
         * ----------------------------------------------------------------
         * Sets or returns the innerHTML of an element, this will not override the existing one. If you want 
         * to override, the call truncate first el.truncate().html(new)
         * ----------------------------------------------------------------
         * @param {*} value Supply this value to set innerHTML
         * @see https://www.247-dev.com/projects/jquery-lite/doc/html
         */

        html(value){
            return html(jQueryLite(allEl[0]), value);
        },

        /**
         * ----------------------------------------------------------------
         * Sets or returns the innerText of an element, this will not override the existing one. If you want 
         * to override, the call truncate first el.truncate().text(new)
         * ----------------------------------------------------------------
         * @param {*} value Supply this value to set innerText
         * @see https://www.247-dev.com/projects/jquery-lite/doc/text
         */

        text(value){
            return text(jQueryLite(allEl), value);
        },

        /**
         * ----------------------------------------------------------------
         * Hides or shows an element depending on the parameter
         * ----------------------------------------------------------------
         * @param {Boolean} bool if false, will show the element else will hide it. true by default
         * @see https://www.247-dev.com/projects/jquery-lite/doc/hide
         */

        hide(bool){
            allEl.map(el => el.hidden = _boolean(bool));
            return jQueryLite(allEl);
        },

        /**
         * ----------------------------------------------------------------
         * Makes an input required or not
         * ----------------------------------------------------------------
         * @param {Boolean} bool if true, will make it required. true by default
         * @see https://www.247-dev.com/projects/jquery-lite/doc/required
         */

        required(bool){
            allEl.map(el => el.required = _boolean(bool));
            return jQueryLite(allEl);
        },

        /**
         * ----------------------------------------------------------------
         * Marks a select option selected or not
         * ----------------------------------------------------------------
         * @param {Boolean} bool if true, will maark it selected. false by default
         * @see https://www.247-dev.com/projects/jquery-lite/doc/selected
         */

        selected(bool){
            allEl.map(el => el.selected = _boolean(bool, false));
            return jQueryLite(allEl);
        },

        /**
         * ----------------------------------------------------------------
         * focuses an element, best for inputs and textareas
         * ----------------------------------------------------------------
         * @see https://www.247-dev.com/projects/jquery-lite/doc/focus
         */

        focus(){
            allEl[0].focus();
            return jQueryLite(allEl);
        },

        /**
         * ----------------------------------------------------------------
         * Blurs or unfocus an element
         * ----------------------------------------------------------------
         * @see https://www.247-dev.com/projects/jquery-lite/doc/blur
         */

        blur(){
            allEl.map(el => el.blur());
            return jQueryLite(allEl)
        },

        /**
         * ----------------------------------------------------------------
         * Selects an element
         * ----------------------------------------------------------------
         * @see https://www.247-dev.com/projects/jquery-lite/doc/select
         */

        select(){
            allEl.map(el => el.select());
            return jQueryLite(allEl)
        },

        /**
         * ----------------------------------------------------------------
         * Submits a form
         * ----------------------------------------------------------------
         * @see https://www.247-dev.com/projects/jquery-lite/doc/submit
         */

        submit(){
            return allEl.map(el => el.submit());
        },

        /**
         * ----------------------------------------------------------------
         * Plays a media element
         * ----------------------------------------------------------------
         * @returns {Promise}
         * @see https://www.247-dev.com/projects/jquery-lite/doc/play
         */

        play(){
            return Promise.all(allEl.map(el => el.play()));
        },

        /**
         * ----------------------------------------------------------------
         * Pauses a media element
         * ----------------------------------------------------------------
         * @see https://www.247-dev.com/projects/jquery-lite/doc/pause
         */

        pause(){
            return jQueryLite(allEl.map(el => el.pause()));

        },

        /**
         * ----------------------------------------------------------------
         * Adds the current element to the end of another
         * ----------------------------------------------------------------
         * @param {Element} parent The receiving parent
         * @see https://www.247-dev.com/projects/jquery-lite/doc/appendto
         */

        appendTo(parent){
            return jQueryLite(allEl.map(el => appendTo(el, parent)));
        },

        /**
         * ----------------------------------------------------------------
         * Adds the current element to the beginning of another
         * ----------------------------------------------------------------
         * @param {Element} parent The receiving parent
         * @see https://www.247-dev.com/projects/jquery-lite/doc/prependto
         */
        
        prependTo(parent){
            return jQueryLite(allEl.reverse().map(el => prependTo(jQueryLite, el, parent)));
        },

        /**
         * ----------------------------------------------------------------
         * Sets or return the value of a form element
         * ----------------------------------------------------------------
         * @param {*} v The new value if set
         * @see https://www.247-dev.com/projects/jquery-lite/doc/value
         */

        value(v){
            if (_primitive(v, null)) {
                allEl.map(el => el.value = v);
                return jQueryLite(allEl);
            } else
                return allEl[0].value;
        },

        /**
         * ----------------------------------------------------------------
         * Gets all the Siblings of the element
         * ----------------------------------------------------------------
         * @param {Boolean} onlyElements If set to true, will get only Elements else 
         * will get every Sibling including texts, comments etc.
         * @see https://www.247-dev.com/projects/jquery-lite/doc/siblings
         */


        siblings(onlyElements = true){
            let parent = allEl[0].parentElement;
            return queryMethods(onlyElements ? parent.children : parent.childNodes, allEl[0], onlyElements)
        },

        /**
         * ----------------------------------------------------------------
         * Will return the parent Element nth times behind
         * ----------------------------------------------------------------
         * @param {Number} n which parent above the tree should be targeted? 0 = parentElement,
         * 1 = parentElement.parentElement, 2 = parentElement.parentElement.parentElement. etc
         * @see https://www.247-dev.com/projects/jquery-lite/doc/parent
         */

        parent(n){
            return jQueryLite(parent(allEl[0], n));
        },

        /**
         * ----------------------------------------------------------------
         * Returns the style object or property value of the element
         * ----------------------------------------------------------------
         * @param {String} prop This would be the property name. If set, will return its value else, 
         * will return the entire style object.
         * @see https://www.247-dev.com/projects/jquery-lite/doc/getstyle
         */

        getStyle(prop){
            let style = getComputedStyle(allEl[0]);
            if(!prop) return style;
            return style.getPropertyValue(prop);
        },

        /**
         * ----------------------------------------------------------------
         * Gets all the childNodes of the element
         * ----------------------------------------------------------------
         * @param {Boolean} onlyElements If set to true, will get only Elements else 
         * will get every childNode including texts, comments etc.
         * @see https://www.247-dev.com/projects/jquery-lite/doc/children
         */

        children(onlyElements = true){
            return queryMethods(onlyElements ? allEl[0].children : allEl[0].childNodes)

        },

        /**
         * ----------------------------------------------------------------
         * Returns a particular childNode by its index
         * ----------------------------------------------------------------
         * @param {Number} index index of the node
         * @param {Boolean} onlyElements If set to true, will verify only amongs Elements else 
         * will check every childNode including texts, comments etc.
         * @see https://www.247-dev.com/projects/jquery-lite/doc/child
         */

        child (index = 0, onlyElement = true){
            return jQueryLite(allEl[0]).children(onlyElement).get(index);
        },

        /**
         * ----------------------------------------------------------------
         * Returns true if the element is visible within the viewport else false
         * ----------------------------------------------------------------
         * @param {Number} offsetX Optional x value to ignore during calculation.
         * @param {Number} offsetY Optional y value to ignore during calculation.
         * @see https://www.247-dev.com/projects/jquery-lite/doc/isvisible
         */

        isVisible(offsetX = 0, offsetY = 0){
            let bc = allEl[0].getBoundingClientRect();
            return (bc.bottom > 0 && bc.top < (innerHeight + offsetY)) && (bc.right > 0 && bc.left < (innerWidth + offsetX));
        },

        /**
         * ----------------------------------------------------------------
         * Gets all the decendants of the element that meet a particular criteria
         * ----------------------------------------------------------------
         * @param {String} criteria should querySelectorAll argument
         * @see https://www.247-dev.com/projects/jquery-lite/doc/query
         */

        query(criteria){
            return queryMethods([].slice.call(allEl[0].querySelectorAll(criteria)));
        },

        /**
         * ----------------------------------------------------------------
         * Gets the first decendants of the element that meet a particular criteria
         * ----------------------------------------------------------------
         * @param {String} criteria should querySelector argument
         * @see https://www.247-dev.com/projects/jquery-lite/doc/find
         */

        find(criteria) {
            return jQueryLite(allEl[0].querySelector(criteria));
        },

        /**
         * Returns the exact height in pixels of the element including scrollbar height;
         */

        height(){
           if(isDocOrWin(allEl[0])) return innerHeight;
              return allEl[0].getBoundingClientRect().height;
        },

        /**
         * Returns the exact width in pixels of the element including scrollbar width;
         */
        
        width(){
            if(isDocOrWin(allEl[0])) return innerWidth;
              return allEl[0].getBoundingClientRect().width
        },
        offsetHeight(){
            return docWinProp(allEl[0], 'offsetHeight');
        },
        offsetWidth(){
            return docWinProp(allEl[0], 'offsetWidth');
        },
        offsetTop(){
            return docWinProp(allEl[0], 'offsetTop');
        },
        offsetLeft(){
            return docWinProp(allEl[0], 'offsetLeft');
        },
        clientHeight(){
            return docWinProp(allEl[0], 'clientHeight');
        },
        clientWidth(){
            return docWinProp(allEl[0], 'clientWidth');
        },
        scrollHeight() {
            return docWinProp(allEl[0], 'scrollHeight');
        },
        scrollWidth(){
            return docWinProp(allEl[0], 'scrollWidth');
        },
        scrollLeft(){
            return docWinProp(allEl[0], 'scrollLeft');
        },
        scrollTop(){
            return docWinProp(allEl[0], 'scrollTop');
        },
        scrollIntoView(...args){
            return  isDocOrWin(allEl[0]) ? docEl(allEl[0]).scrollIntoView(...args) : allEl[0].scrollIntoView(...args);
        },
        scrollTo(...args){
           return  isDocOrWin(allEl[0]) ? window.scrollTo(...args) : allEl[0].scrollTo(...args);
        },
        scrollY(){
            return allEl[0].scrollY;
        },
        scrollX() {
            return allEl[0].scrollX;
        },
        tagName(){
            return allEl[0].tagName
        },
        getBoundingClientRect(...args){
            return isDocOrWin(allEl[0]) ? docEl(allEl[0]).getBoundingClientRect(...args) : allEl[0].getBoundingClientRect(...args);
        },
    
        /**
         * Return the plain HTMLElement(s) in an array because we may be 
         * dealing with more than one element at once
         * @param {Number} n the position of the element you want to retrieve as HTMLElement. It is adviced 
         * to call plain without any parameter and then use array methods to handle it.
         * @returns {[HTMLElement]}
         */

        plain(n){
            return typeof n === 'number' ? allEl[Math.abs(n)] : allEl;
        },

        /**
         * Filters the collection of element. Accepts a function, string and number, please 
         * the link attached
         * @param {*} exclude What to exclude
         * @see https://www.247-dev.com/projects/jquerylite/doc/exclude
         */

        except(query){
            return queryMethods(allEl).except(query);
        },

        /**
         * Because you may be handling more than an element at once, use this method to handle a
         * a particular one by index
         * the link attached
         * @param {Number} n The index of the element to work with, if not provided, the entire elements 
         * are returned untouched
         * @see https://www.247-dev.com/projects/jquerylite/doc/get
         */  

        get(n){
            return jQueryLite(typeof n === 'number' ? allEl[n] : allEl);
        },

        /**
         * runs a function for each of the elements you are working with. Use it as Array.forEach
         * @param {Function} fn callback
         */

        forEach(fn){
            allEl.forEach((el, index, array) => fn(jQueryLite(el), index, array.map(jQueryLite)))
            return jQueryLite(allEl);
        }
    }
}


jQueryLite.isJQL = el => isJQL(el);

/**
 * Create dynamic Bootstrap 5 modal
 * @param {Element | Text} body
 * @param {{
 *  header: Element | Primitive,
 *  footer: Element | Primitive,
 *  background: String,
 *  backdrop: Boolean | 'static',
 *  focus: Boolean,
 *  keyboard: Boolean,
 *  parent: Element,
 *  borderLine: String,
 *  innerScroll: Boolean
 * }} props 
 */

jQueryLite.modal = function(body, props){
    return Modal(jQueryLite, body, _obj(props, {body: props ? props : null}));
};

/**
 * ----------------------------------------------------------------
 * Generates a bootstrap 5 offCanvas
 * ----------------------------------------------------------------
 * @param {Text | Element} body Main content of the offCanvas
 * 
 * @param {{ 
 * title: Text | Element,
 * position: 'top' | 'right' | 'bottom' | 'left',
 * backdrop: Boolean,
 * keyboard: Boolean,
 * scroll: Boolean,
 * parent: Element,
 * background: String,
 * border: String
 * }} props
 * 
 * @property movepoint Can only be dragged from this point if set
 * @property onMove Called immidiately the element is moved
 * @property onMoving Called when the element is moving
 * @property onMoved Called when the element stops after moving
 * @condition The element must have a position property set to 'absolute' or 'fixed'. 
 * You have to set this value by yourself since the method won't temper with your styling
 * @see https://www.247-dev.com/projects/jquerylite/doc/offcanvas
 */

jQueryLite.offcanvas = (body, props) => OffCanvas(jQueryLite, body, _obj(props));

/**
 * Creates a custom alert using Bootstrap 5
 * @param {String} message an information to the user. Can be a string or Element.
 * @param {String | Element} button  Optional button to override default. Can be text or element.
 * 
 * @see https://www.247-dev.com/projects/jqlite/docs/alert
 */
 
jQueryLite.alert = (message, button) => Alert(jQueryLite, message, button);

/**
 * Creates a custom confirm box using Bootstrap 5
 * @param {String} message
 * 
 * @param {{
 * title: String,
 * buttons: [{id: Number | String, value:  String | Element}],
 * }} props a confirmation message for the user or an object containing it.
 * 
 * @property message, a confirmation message for the user
 * @property title, The title of the confirm box
 * @property buttons, All the action buttons and their ids.. optional
 * 
 * @see https://www.247-dev.com/projects/jqlite/docs/confirm
 */
 
jQueryLite.confirm = (message, props) => Confirm(jQueryLite, message, _obj(props));

/**
 * Creates a custom propmt using Bootstrap 5
 * @param {String} message
 * 
 * @param {Text | {
 * type: 'color' | 'date' | 'datetime-local' | 'email' | 'file' | 'month' | 'number' | 'password' | 'range' | 'tel' | 'text' | 'time' | 'url' | 'week',
 * initial: String,
 * okText: String,
 * cancelText: String
 * }} props an information to the user about the input.
 * 
 * @param {String} initial, Only considered if typof message is String.
 * 
 * @property type, The type of input. Should be an HTMlInput type
 * @property okTect, text to display in the Ok button.
 * @property cancelText, text to display in the cancel button.
 * 
 * @see https://www.247-dev.com/projects/jqlite/docs/prompt
 */
 
jQueryLite.prompt = (message, props) => Prompt(jQueryLite, message, _obj(props, {initial: props}));

/**
 * ----------------------------------------------------------------
 * Dynamically generate Bootstrap 5 Form from one to infinite number
 * of fields. Lunch the Form on a modal or append it to a particular 
 * element.
 * ----------------------------------------------------------------
 * 
 * @param {{
 * title: Text | Element,
 * description: Text | Element
 * attributes: {
 *     method: 'post' | 'get' | 'put' | 'delete',
 *     action: URL,
 *     enctype: 'application/x-www-form-urlencoded' | 'multipart/form-data' | 'text/plain'
 * },
 * extraButtons: [Element | Text],
 * submitText: String
 * }} info
 * @see https://www.247-dev.com/projects/jquerylite/doc/form
 */

jQueryLite.form = info => new Form(jQueryLite, info);

/**
 * ----------------------------------------------------------------
 * Dynamically generate and mount Bootstrap 5 Toasts anywhere on the viewport
 * ----------------------------------------------------------------
 * @param {String} message
 * @param {{
 * position: 'tl' | 'tc' | 'tr' | 'ml' | 'mc' | 'mr' | 'bl' | 'bc' | 'br', 
 * title: String, 
 * link: {url: String, name: String} | String, 
 * autohide: Boolean, 
 * animation: Boolean, 
 * delay: Number
 * }} props
 * 
 * @property position, tl = Top left, tc = Top center, tr = Top right, ml = Middle left, mc = Middle center, mr = Middle right, bl = Bottom left, bc = Bottom center, br = Bottom right.
 * @property message The main message you are toasting to the user, must not be too long.
 * @property link, With this set, a button will be added to take take the user to the link provided if clicked. To give link a custom name, the supply an object with url and name properties
 * @property autodelete, Tells whether to delete the toast automatically after a period of time.
 * @property animation, tells whether to animate the appearance and disappearance of a toast or not.
 * @property delay, how long before a toast gets automatically deleted.
 * @see https://www.247-dev.com/projects/jquerylite/doc/toast
 */

jQueryLite.toast = (message, props) => Toast(jQueryLite, message, _obj(props));

  
 /**
  * ----------------------------------------------------------------
  * Dynamically generate a rectangular movable window on the viewport with minimize, maximize and close buttons.
  * ----------------------------------------------------------------
  * @param {HTMLElement | JQL | Primitive} content
  * @param {{
  * border: CSS,
  * height: Number,
  * width: Number,
  * title: String,
  * theme: 'windows' | 'mac',
  * themeColor: CSS
  * } | Number} props All the properties of the window would be extracted from here.
  * 

  * @property title, The title of the window
  * @property height, The height of the window
  * @property width, the width of the window
  * @property border, Should be CSS border property value.
  * @property theme, How the window should appear like. Ms Windows' or Apple Mac's
  * @property themeColor, Background color of the header. Choose a color to match the text color which is dark.
  * @see https://www.247-dev.com/projects/jquerylite/doc/window
  */

jQueryLite.window = (content, props) => Window(jQueryLite, content, _obj(props));

/**
 * Dynamically generate a nested collapsible element
 * @param {[{
 * href: URL,
 * id: String,
 * label: String,
 * children: items
 *  }]} items
 * 
 * @param {{
 * grid: 'vertical' | 'both' | 'none',
 * active: URL, 
 * linkIcon: SVGElement,
 * onclick: Function
 *  }} props 
 * @see https://www.247-dev.com/projects/jquerylite/doc/nestedcollapse
 */

jQueryLite.nestedCollapse = (items, props) => nestedCollapse(jQueryLite, items, _obj(props));

/**
 * Generate a fresh Bootstrap 5 carousel without having to write any HTML tags 
 * @param {[Element | Text | {
 * item: Element | Text | URL,
 * interval: Number | Boolean,
 * type: 'picture' | 'video' | 'other',
 * autoplay: Boolean,
 * controls: Boolean
 * }]} items
 * 
 * @param {{
 * active: Number,
 * autostart: Boolean,
 * speed: Number,
 * transition: 'fade' | 'slide',
 * interval: Number | Boolean,
 * arrows: Boolean,
 * indicators: Boolean,
 * hover: Boolean,
 * keyboard: Boolean,
 * touch: Boolean,
 * recycle: Boolean,
 * theme: 'dark' | 'light'
 * }} props 
 * @see https://www.247-dev.com/projects/jquerylite/doc/carousel
 */

jQueryLite.carousel = (items, props) => Carousel(jQueryLite, items, _obj(props));

/**
 * ----------------------------------------------------------------
 * Generates a manual slider
 * ----------------------------------------------------------------
 * @param {[{
 * title: Text | Element,
 * body: Element | Text | {
 * type: 'video' | 'picture' | 'other',
 * item: Element | Text | URL,
 * interval: Number,
 * controls: Boolean,
 * autoplay: Boolean,
 * }
 * }]} slides
 * 
 * @param {{
 * footer: Text | Element,
 * speed: Number,
 * alignHeader: 'left' | 'center' | 'right',
 * alignFooter: 'left' | 'center' | 'right',
 * breakHead: Boolean,
 * breakFoot: Boolean,
 * touch: Boolean
 * }} props
 * @see https://www.247-dev.com/projects/jquerylite/doc/section
 */

jQueryLite.section = (slides, props) => Section(jQueryLite, slides, _obj(props));

/**
 * 
 * @param {Text | Element} body 
 * @param {{
 * header: Text | Element,
 * footer: Text | Element
 * }} props 
 * @see https://www.247-dev.com/projects/jquerylite/doc/card
 */

jQueryLite.card = (body, props) => jQueryLite(Card(jQueryLite, body, _obj(props)))

/**
 * Copies texts to clipboard
 * @param {String} text, text to be copied
 * 
 * @returns {Promise}
 * @see https://www.247-dev.com/projects/jquerylite/doc/copy
 */

jQueryLite.copy = async(text) => {
    try {
        return navigator.clipboard.writeText(text);
    } catch (e) {
      return new Promise(res => {
        let input = jQueryLite('<input/>').value(text).plain()
        jQueryLite(document.body).addChild(input);
        input.select();
        input.setSelectionRange(0, 99999);
        document.execCommand("copy");
        jQueryLite(input).drop();
        res()
    })
    }
}
