import {sleep} from "../../utils/lib.js";

export default function({ el, transitionDuration: td, transitionDistance: _td, criteria, transitionType: tt }) {
    td = typeof td === 'number' ? td : 1000;
    _td = typeof _td === 'number' ? _td : 200;
    tt = (typeof tt === 'string' && ['to-top', 'to-bottom', 'to-left', 'to-right'].includes(tt)) ? tt : 'to-top';
    let x, y;

    switch (tt) {
        case 'to-top':
            x = 0;
            y = _td;
            break;
        case 'to-bottom':
            x = 0;
            y = -_td;
            break;
        case 'to-left':
            x = _td;
            y = 0;
            break;
        case 'to-right':
            x = -_td;
            y = 0;
    }
    let ch;
    if (criteria && typeof criteria === 'string')
        ch = el.query(criteria);
    else
        ch = el.children().get();

    const init = () => ch.forEach(async(c) => {
        let transform = c.getStyle('transform');
        let opacity = c.getStyle('opacity');

        c.attr({ 'gentle-scroll-initial-styles': JSON.stringify({ transform, opacity }) });
        c.attr({ 'gentle-scroll-initial-transition': c.getStyle('transition') });
        c.style({ opacity: '0', transform: `translate(${x}px, ${y}px)` });
    });


    const runGentle = () => {
        ch.forEach(async(c) => {
            if (c.isVisible(x, y)) {
                c.style(JSON.parse(c.attr('gentle-scroll-initial-styles')))
                await sleep(td);
                c.style({ transition: c.attr('gentle-scroll-initial-transition') })
            }
        });
    }

    init();
    ch.forEach(c => c.style({ transition: `${td}ms all` }));
    runGentle();
    el.on('scroll', runGentle);
}